"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("./types");
let optimisticIdCounter = -1;
/**
 * Generates an id by counting.
 * @returns {string}
 */
function generateId() {
    optimisticIdCounter++;
    return optimisticIdCounter.toString();
}
let defaultTimeout = 20000;
function setDefaultTimeout(timeout) {
    defaultTimeout = timeout;
}
exports.setDefaultTimeout = setDefaultTimeout;
function getDefaultTimeout() {
    return defaultTimeout;
}
exports.getDefaultTimeout = getDefaultTimeout;
let defaultTimeoutHandler = (optimisticId, commit, rollback) => {
    console.log('Neither commit or rollback was called for optimism with ID \'' + optimisticId + "\'. Leaving" +
        " the state pending for a longer time will result in an increased size of the state and to " +
        "it being incomprehensible.");
    rollback();
};
function setDefaultTimeoutHandler(func) {
    defaultTimeoutHandler = func;
}
exports.setDefaultTimeoutHandler = setDefaultTimeoutHandler;
function getDefaultTimeoutHandler() {
    return defaultTimeoutHandler;
}
exports.getDefaultTimeoutHandler = getDefaultTimeoutHandler;
/**
 * Simplifies the usage of redux-optimism. By calling this function with the dispatch function of your store,
 * redux-optimism will dispatch an 'INIT_OPTIMISM' action with a newly created ID and give you back three functions:
 *  * makeOptimistic
 *  * commit
 *  * rollback
 * With makeOptimistic(action), you can optimistically dispatch an action. With commit(), the previously executed makeOptimistic
 * actions will be manifested. And with rollback(), the state will be the one, that would have occurred without the
 * optimistically dispatched actions.
 * @param dispatch
 * @returns {{makeOptimistic: (action: FSA) => never, rollback: () => never, timer: number, dispatched: never, commit: () => never}}
 * @param timeout
 * @param handleTimeout
 */
function getOptimism(dispatch, timeout = getDefaultTimeout(), handleTimeout = getDefaultTimeoutHandler()) {
    const optimisticId = generateId();
    const result = dispatch({
        type: types_1.optimisticActions.init,
        payload: { optimisticId: optimisticId }
    });
    function optimistic(action) {
        const modifiedAction = {
            ...action,
            meta: {
                ...action.meta,
                optimisticId: optimisticId
            }
        };
        return dispatch(modifiedAction);
    }
    function commit() {
        clearTimeout(timer);
        return dispatch({ type: types_1.optimisticActions.commit, payload: { optimisticId: optimisticId } });
    }
    function rollback() {
        clearTimeout(timer);
        return dispatch({ type: types_1.optimisticActions.rollback, payload: { optimisticId: optimisticId } });
    }
    let timer = setTimeout(() => {
        handleTimeout(optimisticId, commit, rollback);
    }, timeout);
    return {
        optimistic: optimistic,
        commit: commit,
        rollback: rollback,
        dispatched: result,
        timer: timer // For testing
    };
}
exports.getOptimism = getOptimism;
/**
 * By dispatching this action, the actions, that were dispatched before all active optimisms, will be integrated into a
 * new realistic state. Saves time, when there is an rollback. Should be dispatched in idle time.
 * @type {{type: optimisticActions.cleanup}}
 */
exports.cleanStore = {
    type: types_1.optimisticActions.cleanup
};
//# sourceMappingURL=actions.js.map