"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("./types");
// Add this function -- otherwise testing doesn't work...
Object.fromEntries = (array) => Object.assign({}, ...Array.from(array, ([key, value]) => ({ [key]: value })));
let actionIdCounter = -1;
/**
 * Creates an ID by counting.
 * @returns {string}
 */
function getNextId() {
    actionIdCounter++;
    return actionIdCounter.toString();
}
/**
 * The identity operator.
 * @param x
 * @returns {any}
 */
function identity(x) {
    return x;
}
exports.identity = identity;
/**
 * Map function for objects. The first argument is the object, the second argument is a function that takes a tuple
 * [key, value] and returns a tuple [newKey, newValue]
 * @param object
 * @param mapFunction
 * @returns {any}
 */
function objectMap(object, mapFunction = identity) {
    return Object.fromEntries(Object.entries(object).map(([k, v]) => mapFunction([k, v])));
}
exports.objectMap = objectMap;
/**
 * Filter function for objects. The first argument is the object, the filterFunction takes a tuple [key, value] and
 * returns a boolean.
 * @param object
 * @param filterFunction
 * @returns {any}
 */
function objectFilter(object, filterFunction = () => true) {
    return Object.fromEntries(Object.entries(object).filter(([k, v]) => filterFunction([k, v])));
}
exports.objectFilter = objectFilter;
/**
 * Calculates the intersection between arrays.
 * @param arrays
 * @returns {Array<any>}
 */
function intersect(...arrays) {
    return arrays.reduce((accumulator, current) => {
        return accumulator.filter((x) => current.includes(x));
    });
}
exports.intersect = intersect;
const initialState = {
    optimism: {
        actions: {},
        actionsUntilNow: [],
        optimisms: {},
        realisticState: {}
    }
};
/**
 * Enhances a reducer with optimism.
 * @param wrappedReducer
 * @param actionIdGenerator
 */
function makeOptimistic(wrappedReducer, actionIdGenerator = getNextId) {
    function optimisticReducer(previousState = initialState, action) {
        const actionId = actionIdGenerator();
        let optimisticState;
        optimisticState = {
            ...previousState,
            optimism: Object.assign({}, initialState.optimism, previousState.optimism)
        };
        let intermediateState;
        switch (action.type) {
            case types_1.optimisticActions.init:
                intermediateState = init(optimisticState, action, actionId);
                break;
            case types_1.optimisticActions.commit:
                intermediateState = commit(optimisticState, action, actionId);
                break;
            case types_1.optimisticActions.rollback:
                intermediateState = rollback(optimisticState, action, actionId, wrappedReducer);
                break;
            /**
             case optimisticActions.cleanup:
             intermediateState = cleanup(optimisticState, action, actionId, wrappedReducer)
             break
             **/
            default:
                intermediateState = defaultReducer(optimisticState, action, actionId, wrappedReducer);
                break;
        }
        // Always do the cheap cleanup
        if (Object.keys(intermediateState.optimism.optimisms).length === 0) {
            return {
                ...intermediateState,
                optimism: {
                    ...intermediateState.optimism,
                    actions: {},
                    actionsUntilNow: [],
                    realisticState: objectFilter(intermediateState, ([key, value]) => key != 'optimism')
                }
            };
        }
        else {
            return intermediateState;
        }
    }
    return optimisticReducer;
}
exports.makeOptimistic = makeOptimistic;
/**
 * Update the state for an makeOptimistic init.
 * @param previousState
 * @param action
 * @param actionId
 * @returns {{[p: string]: any, optimism: {realisticState: any, actionsUntilNow: string[], optimisms: any, actions: {[p: string]: FSA, [p: string]: FSA}}}}
 */
function init(previousState, action, actionId) {
    if (!(action.payload) || action.payload.optimisticId == null) {
        throw Error("INIT_OPTIMISM was dispatched without an optimisticId!");
    }
    const optimisticId = action.payload.optimisticId;
    if (Object.keys(previousState.optimism.optimisms).includes(optimisticId)) {
        throw Error('Can\'t initialize Optimism with ID \'' + optimisticId + '\' -- it has already been initialized.');
    }
    return {
        ...previousState,
        optimism: {
            ...previousState.optimism,
            actions: {
                ...previousState.optimism.actions,
                [actionId]: action
            },
            actionsUntilNow: [
                ...previousState.optimism.actionsUntilNow,
                actionId
            ],
            optimisms: {
                ...objectMap(previousState.optimism.optimisms, ([optimismId, optimism]) => ([optimismId, {
                        ...optimism,
                        subsequentActions: [...optimism.subsequentActions, actionId]
                    }])),
                [optimisticId]: {
                    previousActions: previousState.optimism.actionsUntilNow,
                    subsequentActions: []
                }
            }
        }
    };
}
/**
 * Update the state in case of an makeOptimistic commit.
 * @param previousState
 * @param action
 * @param actionId
 * @returns {{[p: string]: any; optimism: {realisticState: any; actionsUntilNow: string[]; optimisms: any; actions: {[p: string]: FSA; [p: string]: FSA}}}}
 */
function commit(previousState, action, actionId) {
    if (!(action.payload) || action.payload.optimisticId == null) {
        throw Error("COMMIT_OPTIMISM was dispatched without an optimisticId!");
    }
    const optimisticId = action.payload.optimisticId;
    if (!(Object.keys(previousState.optimism.optimisms).includes(optimisticId))) {
        throw Error('Can\'t commit optimism with ID \'' + optimisticId + '\' -- either it wasn\'t initialized' +
            'or it\'s already committed/rolled back.');
    }
    return {
        ...previousState,
        optimism: {
            ...previousState.optimism,
            actions: {
                ...previousState.optimism.actions,
                [actionId]: action
            },
            actionsUntilNow: [
                ...previousState.optimism.actionsUntilNow,
                actionId
            ],
            optimisms: {
                ...objectFilter(previousState.optimism.optimisms, ([optimismId, optimism]) => {
                    return optimismId != optimisticId;
                })
            }
        }
    };
}
/**
 * Perform the rollback of an optimism.
 * @param previousState
 * @param action
 * @param actionId
 * @param wrappedReducer
 * @returns {{payload?: any; optimism: {realisticState: any; actionsUntilNow: string[]; optimisms: any; actions: {[p: string]: FSA; [p: string]: FSA}}; meta?: any; type: string; error?: boolean | null | undefined}}
 */
function rollback(previousState, action, actionId, wrappedReducer) {
    if (!(action.payload) || action.payload.optimisticId == null) {
        throw Error("ROLLBACK_OPTIMISM was dispatched without an optimisticId!");
    }
    const optimisticId = action.payload.optimisticId;
    if (!(Object.keys(previousState.optimism.optimisms).includes(optimisticId))) {
        throw Error('Can\'t rollback optimism with ID \'' + optimisticId + '\' -- either it wasn\'t initialized' +
            'or it\'s already committed/rolled back.');
    }
    const actionIds = previousState.optimism.optimisms[optimisticId].previousActions
        .concat(previousState.optimism.optimisms[optimisticId].subsequentActions);
    const actions = actionIds.map((actionId) => previousState.optimism.actions[actionId]);
    const integratedState = {
        ...(actions.reduce((accumulatedState, action) => {
            return makeOptimistic(wrappedReducer)(accumulatedState, action);
        }, previousState.optimism.realisticState)),
        optimism: previousState.optimism
    };
    return {
        ...integratedState,
        optimism: {
            ...integratedState.optimism,
            actions: {
                ...integratedState.optimism.actions,
                [actionId]: action
            },
            actionsUntilNow: [
                ...integratedState.optimism.actionsUntilNow,
                actionId
            ],
            optimisms: {
                ...objectFilter({
                    ...objectMap(integratedState.optimism.optimisms, ([optimismId, optimism]) => {
                        if (optimismId == optimisticId) {
                            return [optimismId, optimism];
                        }
                        else {
                            return [optimismId, {
                                    ...optimism,
                                    subsequentActions: [...optimism.subsequentActions, actionId]
                                }];
                        }
                    })
                }, ([key, value]) => {
                    return key != optimisticId;
                })
            }
        }
    };
}
/**
 * Reduce every other action.
 * @param previousState
 * @param action
 * @param actionId
 * @param wrappedReducer
 * @returns {any}
 */
function defaultReducer(previousState, action, actionId, wrappedReducer) {
    if (types_1.isOptimistic(action)) {
        let optimisticId = action.meta.optimisticId;
        if (!(previousState.optimism) || !(Object.keys(previousState.optimism.optimisms).includes(optimisticId))) {
            throw Error('Can\'t optimistically apply action \'' + action.type + '\' -- the optimism with the' +
                ' ID \'' + optimisticId + '\' was either never initialized or was already committed/rolled back.');
        }
    }
    const nextState = wrappedReducer(previousState, action);
    const nextStateOptimistic = {
        ...nextState,
        optimism: Object.assign({}, initialState, nextState.optimism)
    };
    return {
        ...nextStateOptimistic,
        optimism: {
            ...nextStateOptimistic.optimism,
            actions: {
                ...nextStateOptimistic.optimism.actions,
                [actionId]: action
            },
            actionsUntilNow: [
                ...nextStateOptimistic.optimism.actionsUntilNow,
                actionId
            ],
            optimisms: {
                ...objectMap(nextStateOptimistic.optimism.optimisms, ([optimismId, optimism]) => {
                    if (types_1.isOptimistic(action) && (action.meta.optimisticId === optimismId)) {
                        return [optimismId, optimism];
                    }
                    else {
                        return [optimismId, {
                                ...optimism,
                                subsequentActions: [...optimism.subsequentActions, actionId]
                            }];
                    }
                })
            }
        }
    };
}
/**
 * Cleanup the state.
 * @param previousState
 * @param action
 * @param actionId
 * @param wrappedReducer
 * @returns {any}
 */
/**
function cleanup(
 previousState: optimisticState,
 action: FSA,
 actionId: string,
 wrappedReducer: Reducer): optimisticState {

    if (Object.keys(previousState.optimism.optimisms).length === 0) {
        return {
            ...previousState,
            optimism: {
                ...previousState.optimism,
                actions: {},
                actionsUntilNow: [],
                realisticState: objectFilter(
                    previousState,
                    ([key, value]: [string, any]) => key != 'optimism'
                )
            }
        }

    } else {
        const commonActions = intersect(
            ...Object.values(previousState.optimism.optimisms).map(
                (optimism: any) => optimism.previousActions)
        )

        const newRealisticState = commonActions.reduce(
            (accumulatedState: any, action: FSA) => {
                return makeOptimistic(wrappedReducer)(accumulatedState, action)
            },
            previousState.optimism.realisticState)

        return {
            ...previousState,
            optimism: {

                ...previousState.optimism,

                actions: objectFilter(
                    previousState.optimism.actions,
                    ([actionId, action]: [string, FSA]) => !commonActions.includes(actionId)
                ),

                actionsUntilNow: previousState.optimism.actionsUntilNow.filter(
                    (actionId: string) => !commonActions.includes(actionId)
                ),

                optimisms: objectMap(
                    previousState.optimism.optimisms,
                    ([optimismId, optimism]: [string, optimism]) => [optimismId, {
                        ...optimism,
                        previousActions: optimism.previousActions.filter(
                            (actionId: string) => !commonActions.includes(actionId)
                        )
                    }]),

                realisticState: objectFilter(
                    newRealisticState,
                    ([key, value]: [string, any]) => key != 'optimism'
                )
            }
        }
    }
}
 */
//# sourceMappingURL=enhancer.js.map